import cffi

ffi = cffi.FFI()
ffi.set_source(
    "kociemba.ckociembawrapper",
    """
    #include "search.h"

    char* solve(char *cubestring, char *patternstring, char *cache_dir, int max_depth)
    {
        char patternized[64];
        char *sol;

        if (patternstring) {
            patternize(cubestring, patternstring, patternized);
            cubestring = patternized;
        }

        sol = solution(
            cubestring,
            max_depth,
            1000,
            0,
            cache_dir
        );
        return sol;
    }
    """,
    include_dirs=['kociemba/ckociemba/include'],
    sources=[
        'kociemba/ckociemba/prunetable_helpers.c',
        'kociemba/ckociemba/coordcube.c',
        'kociemba/ckociemba/cubiecube.c',
        'kociemba/ckociemba/facecube.c',
        'kociemba/ckociemba/search.c'
    ],
    extra_compile_args=['-std=c99', '-O3', '-D_XOPEN_SOURCE=700']
)

ffi.cdef("char* solve(char *cubestring, char *patternstring, char *cache_dir, int max_depth);")

if __name__ == "__main__":
    ffi.compile(verbose=True)



import warnings


class SlowContextWarning(Warning):
    pass

try:
    import os
    from .ckociembawrapper import ffi, lib
    cache_dir = os.path.join(os.path.dirname(__file__), 'cprunetables')

    def _solve(cube, pattern, max_depth=24):
        pattern_utf8 = pattern.encode('utf-8') if pattern is not None else ffi.NULL
        res = lib.solve(cube.encode('utf-8'), pattern_utf8, cache_dir.encode('utf-8'), max_depth)
        if res != ffi.NULL:
            return ffi.string(res).strip().decode('utf-8')
        else:
            raise ValueError('Error. Probably cubestring is invalid')
except ImportError as e:
    print(e)
    warnings.warn("Native version of the package is not available. "
                  "We have to fallback to pure-Python implementation of "
                  "the algorithm, which is usually many times slower. "
                  "If performance is important to you, check official "
                  "project page for a native implementation: "
                  "https://github.com/muodov/kociemba",
                  SlowContextWarning)
    from .pykociemba import search

    def _solve(cube, pattern, max_depth=24):
        errors = {
            'Error 1': 'There is not exactly one facelet of each colour',
            'Error 2': 'Not all 12 edges exist exactly once',
            'Error 3': 'Flip error: One edge has to be flipped',
            'Error 4': 'Not all corners exist exactly once',
            'Error 5': 'Twist error: One corner has to be twisted',
            'Error 6': 'Parity error: Two corners or two edges have to be exchanged',
            'Error 7': 'No solution exists for the given maxDepth',
            'Error 8': 'Timeout, no solution within given time'
        }
        if pattern is not None:
            cube = search.patternize(cube, pattern)
        res = search.Search().solution(cube, max_depth, 1000, False).strip()
        if res in errors:
            raise ValueError(errors[res])
        else:
            return res


def solve(cubestring, patternstring=None, max_depth=24):
    """
    Solve a Rubik's cube using two-phase algorithm.

    >>> solve("BBURUDBFUFFFRRFUUFLULUFUDLRRDBBDBDBLUDDFLLRRBRLLLBRDDF")
    "B U' L' D' R' D' L2 D' L F' L' D F2 R2 U R2 B2 U2 L2 F2 D'"

    >>> kociemba.solve('FLBUULFFLFDURRDBUBUUDDFFBRDDBLRDRFLLRLRULFUDRRBDBBBUFL', 'BBURUDBFUFFFRRFUUFLULUFUDLRRDBBDBDBLUDDFLLRRBRLLLBRDDF')
    u"R' D2 R' U2 R F2 D B2 U' R F' U R2 D L2 D' B2 R2 B2 U' B2"
    """

    return _solve(cubestring, patternstring, max_depth)

__all__ = ['solve']



from builtins import range
import logging
import os.path

try:
    import cPickle
except ImportError:
    import pickle as cPickle

from .cubiecube import CubieCube, moveCube, getURtoDF

log = logging.getLogger(__name__)

cache_dir = os.path.join(os.path.dirname(__file__), 'prunetables')


def setPruning(table, index, value):
    """Set pruning value in table. Two values are stored in one byte."""
    if ((index & 1) == 0):
        table[index // 2] &= 0xf0 | value
    else:
        table[index // 2] &= 0x0f | (value << 4)
    # table[index] = value & 0xf


def getPruning(table, index):
    """Extract pruning value"""

    if ((index & 1) == 0):
        res = table[index // 2] & 0x0f
    else:
        res = (table[index // 2] & 0xf0) >> 4
    return res
    # return table[index] & 0xf


def load_cachetable(name):
    obj = None
    try:
        with open(os.path.join(cache_dir, name + '.pkl'), 'rb') as f:
            obj = cPickle.load(f)
    except IOError as e:
        log.warning('could not read cache for %s: %s. Recalculating it...', name, e)
    return obj


def dump_cachetable(obj, name):
    with open(os.path.join(cache_dir, name + '.pkl'), 'wb') as f:
        cPickle.dump(obj, f)


class CoordCube(object):
    """Representation of the cube on the coordinate level"""

    N_TWIST = 2187  # 3^7 possible corner orientations
    N_FLIP = 2048   # 2^11 possible edge flips
    N_SLICE1 = 495  # 12 choose 4 possible positions of FR,FL,BL,BR edges
    N_SLICE2 = 24   # 4! permutations of FR,FL,BL,BR edges in phase2
    N_PARITY = 2    # 2 possible corner parities
    N_URFtoDLF = 20160  # 8!/(8-6)! permutation of URF,UFL,ULB,UBR,DFR,DLF corners
    N_FRtoBR = 11880    # 12!/(12-4)! permutation of FR,FL,BL,BR edges
    N_URtoUL = 1320     # 12!/(12-3)! permutation of UR,UF,UL edges
    N_UBtoDF = 1320     # 12!/(12-3)! permutation of UB,DR,DF edges
    N_URtoDF = 20160    # 8!/(8-6)! permutation of UR,UF,UL,UB,DR,DF edges in phase2

    N_URFtoDLB = 40320  # 8! permutations of the corners
    N_URtoBR = 479001600    # 8! permutations of the corners

    N_MOVE = 18

    # All coordinates are 0 for a solved cube except for UBtoDF, which is 114
    # short twist
    # short flip
    # short parity
    # short FRtoBR
    # short URFtoDLF
    # short URtoUL
    # short UBtoDF
    # int URtoDF

    def __init__(self, c):
        """
        Generate a CoordCube from a CubieCube

        c - CubieCube instance
        """

        self.twist = c.getTwist()
        self.flip = c.getFlip()
        self.parity = c.cornerParity()
        self.FRtoBR = c.getFRtoBR()
        self.URFtoDLF = c.getURFtoDLF()
        self.URtoUL = c.getURtoUL()
        self.UBtoDF = c.getUBtoDF()
        self.URtoDF = c.getURtoDF()     # only needed in phase2

    def move(self, m):
        """
        A move on the coordinate level

        m - int
        """
        self.twist = self.twistMove[self.twist][m]
        self.flip = self.flipMove[self.flip][m]
        self.parity = self.parityMove[self.parity][m]
        self.FRtoBR = self.FRtoBR_Move[self.FRtoBR][m]
        self.URFtoDLF = self.URFtoDLF_Move[self.URFtoDLF][m]
        self.URtoUL = self.URtoUL_Move[self.URtoUL][m]
        self.UBtoDF = self.UBtoDF_Move[self.UBtoDF][m]
        if (self.URtoUL < 336 and self.UBtoDF < 336):
            # updated only if UR,UF,UL,UB,DR,DF
            # are not in UD-slice
            self.URtoDF = self.MergeURtoULandUBtoDF[self.URtoUL][self.UBtoDF]

    # ******************************************Phase 1 move tables*****************************************************

    log.debug('Preparing move table for the twists of the corners...')
    # Move table for the twists of the corners
    # twist < 2187 in phase 2.
    # twist = 0 in phase 2.
    twistMove = load_cachetable('twistMove')
    if not twistMove:
        twistMove = [[0] * N_MOVE for i in range(N_TWIST)]   # new short[N_TWIST][N_MOVE]
        a = CubieCube()
        for i in range(N_TWIST):
            a.setTwist(i)
            for j in range(6):
                for k in range(3):
                    a.cornerMultiply(moveCube[j])
                    twistMove[i][3 * j + k] = a.getTwist()
                a.cornerMultiply(moveCube[j])   # 4. faceturn restores
                # a
        dump_cachetable(twistMove, 'twistMove')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Move table for the flips of the edges
    # flip < 2048 in phase 1
    # flip = 0 in phase 2.
    log.debug('Preparing move table for the flips of the edges')

    flipMove = load_cachetable('flipMove')
    if not flipMove:
        flipMove = [[0] * N_MOVE for i in range(N_FLIP)]     # new short[N_FLIP][N_MOVE]
        a = CubieCube()
        for i in range(N_FLIP):
            a.setFlip(i)
            for j in range(6):
                for k in range(3):
                    a.edgeMultiply(moveCube[j])
                    flipMove[i][3 * j + k] = a.getFlip()
                a.edgeMultiply(moveCube[j])
                # a
        dump_cachetable(flipMove, 'flipMove')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Parity of the corner permutation. This is the same as the parity for the edge permutation of a valid cube.
    # parity has values 0 and 1
    parityMove = [
        [1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1],
        [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
    ]

    # ***********************************Phase 1 and 2 movetable********************************************************
    log.debug('Preparing move table for the four UD-slice edges FR, FL, Bl and BR')
    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Move table for the four UD-slice edges FR, FL, Bl and BR
    # FRtoBRMove < 11880 in phase 1
    # FRtoBRMove < 24 in phase 2
    # FRtoBRMove = 0 for solved cube

    FRtoBR_Move = load_cachetable('FRtoBR_Move')
    if not FRtoBR_Move:
        FRtoBR_Move = [[0] * N_MOVE for i in range(N_FRtoBR)]    # new short[N_FRtoBR][N_MOVE]
        a = CubieCube()
        for i in range(N_FRtoBR):
            a.setFRtoBR(i)
            for j in range(6):
                for k in range(3):
                    a.edgeMultiply(moveCube[j])
                    FRtoBR_Move[i][3 * j + k] = a.getFRtoBR()
                a.edgeMultiply(moveCube[j])
        dump_cachetable(FRtoBR_Move, 'FRtoBR_Move')

    # *******************************************Phase 1 and 2 movetable************************************************

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Move table for permutation of six corners. The positions of the DBL and DRB corners are determined by the parity.
    # URFtoDLF < 20160 in phase 1
    # URFtoDLF < 20160 in phase 2
    # URFtoDLF = 0 for solved cube.
    log.debug('Preparing move table for permutation of six corners. The positions of the DBL and DRB corners are determined by the parity.')
    URFtoDLF_Move = load_cachetable('URFtoDLF_Move')
    if not URFtoDLF_Move:
        URFtoDLF_Move = [[0] * N_MOVE for i in range(N_URFtoDLF)]    # new short[N_URFtoDLF][N_MOVE]
        a = CubieCube()
        for i in range(N_URFtoDLF):
            a.setURFtoDLF(i)
            for j in range(6):
                for k in range(3):
                    a.cornerMultiply(moveCube[j])
                    URFtoDLF_Move[i][3 * j + k] = a.getURFtoDLF()
                a.cornerMultiply(moveCube[j])
        dump_cachetable(URFtoDLF_Move, 'URFtoDLF_Move')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Move table for the permutation of six U-face and D-face edges in phase2. The positions of the DL and DB edges are
    # determined by the parity.
    # URtoDF < 665280 in phase 1
    # URtoDF < 20160 in phase 2
    # URtoDF = 0 for solved cube.
    log.debug('Preparing move table for the permutation of six U-face and D-face edges in phase2. The positions of the DL and DB edges are')
    URtoDF_Move = load_cachetable('URtoDF_Move')
    if not URtoDF_Move:
        URtoDF_Move = [[0] * N_MOVE for i in range(N_URtoDF)]    # new short[N_URtoDF][N_MOVE]
        a = CubieCube()
        for i in range(N_URtoDF):
            a.setURtoDF(i)
            for j in range(6):
                for k in range(3):
                    a.edgeMultiply(moveCube[j])
                    URtoDF_Move[i][3 * j + k] = a.getURtoDF()
                    # Table values are only valid for phase 2 moves!
                    # For phase 1 moves, casting to short is not possible.
                a.edgeMultiply(moveCube[j])
        dump_cachetable(URtoDF_Move, 'URtoDF_Move')

    # **************************helper move tables to compute URtoDF for the beginning of phase2************************
    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Move table for the three edges UR,UF and UL in phase1.
    log.debug('Preparing move table for the three edges UR,UF and UL in phase1.')
    URtoUL_Move = load_cachetable('URtoUL_Move')
    if not URtoUL_Move:
        URtoUL_Move = [[0] * N_MOVE for i in range(N_URtoUL)]    # new short[N_URtoUL][N_MOVE]
        a = CubieCube()
        for i in range(N_URtoUL):
            a.setURtoUL(i)
            for j in range(6):
                for k in range(3):
                    a.edgeMultiply(moveCube[j])
                    URtoUL_Move[i][3 * j + k] = a.getURtoUL()
                a.edgeMultiply(moveCube[j])
        dump_cachetable(URtoUL_Move, 'URtoUL_Move')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Move table for the three edges UB,DR and DF in phase1.
    log.debug('Preparing move table for the three edges UB,DR and DF in phase1.')
    UBtoDF_Move = load_cachetable('UBtoDF_Move')
    if not UBtoDF_Move:
        UBtoDF_Move = [[0] * N_MOVE for i in range(N_UBtoDF)]    # new short[N_UBtoDF][N_MOVE]
        a = CubieCube()
        for i in range(N_UBtoDF):
            a.setUBtoDF(i)
            for j in range(6):
                for k in range(3):
                    a.edgeMultiply(moveCube[j])
                    UBtoDF_Move[i][3 * j + k] = a.getUBtoDF()
                a.edgeMultiply(moveCube[j])
        dump_cachetable(UBtoDF_Move, 'UBtoDF_Move')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Table to merge the coordinates of the UR,UF,UL and UB,DR,DF edges at the beginning of phase2
    log.debug('Preparing table to merge the coordinates of the UR,UF,UL and UB,DR,DF edges at the beginning of phase2')
    MergeURtoULandUBtoDF = load_cachetable('MergeURtoULandUBtoDF')
    if not MergeURtoULandUBtoDF:
        MergeURtoULandUBtoDF = [[0] * 336 for i in range(336)]   # new short[336][336]
        # for i, j <336 the six edges UR,UF,UL,UB,DR,DF are not in the
        # UD-slice and the index is <20160
        for uRtoUL in range(336):
            for uBtoDF in range(336):
                MergeURtoULandUBtoDF[uRtoUL][uBtoDF] = getURtoDF(uRtoUL, uBtoDF)
        dump_cachetable(MergeURtoULandUBtoDF, 'MergeURtoULandUBtoDF')

    # ****************************************Pruning tables for the search*********************************************
    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Pruning table for the permutation of the corners and the UD-slice edges in phase2.
    # The pruning table entries give a lower estimation for the number of moves to reach the solved cube.
    log.debug('Preparing pruning table for the permutation of the corners and the UD-slice edges in phase2.')
    Slice_URFtoDLF_Parity_Prun = load_cachetable('Slice_URFtoDLF_Parity_Prun')
    if not Slice_URFtoDLF_Parity_Prun:
        Slice_URFtoDLF_Parity_Prun = [-1] * (N_SLICE2 * N_URFtoDLF * N_PARITY // 2)     # new byte[N_SLICE2 * N_URFtoDLF * N_PARITY / 2]
        # Slice_URFtoDLF_Parity_Prun = [-1] * (N_SLICE2 * N_URFtoDLF * N_PARITY)
        depth = 0
        setPruning(Slice_URFtoDLF_Parity_Prun, 0, 0)
        done = 1
        while (done != N_SLICE2 * N_URFtoDLF * N_PARITY):
            for i in range(N_SLICE2 * N_URFtoDLF * N_PARITY):
                parity = i % 2
                URFtoDLF = (i // 2) // N_SLICE2
                _slice = (i // 2) % N_SLICE2
                if getPruning(Slice_URFtoDLF_Parity_Prun, i) == depth:
                    for j in range(18):
                        if j in (3, 5, 6, 8, 12, 14, 15, 17):
                            continue
                        else:
                            newSlice = FRtoBR_Move[_slice][j]
                            newURFtoDLF = URFtoDLF_Move[URFtoDLF][j]
                            newParity = parityMove[parity][j]
                            if (getPruning(Slice_URFtoDLF_Parity_Prun, (N_SLICE2 * newURFtoDLF + newSlice) * 2 + newParity) == 0x0f):
                                setPruning(
                                    Slice_URFtoDLF_Parity_Prun,
                                    (N_SLICE2 * newURFtoDLF + newSlice) * 2 + newParity,
                                    (depth + 1) & 0xff
                                )
                                done += 1

            depth += 1
        dump_cachetable(Slice_URFtoDLF_Parity_Prun, 'Slice_URFtoDLF_Parity_Prun')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Pruning table for the permutation of the edges in phase2.
    # The pruning table entries give a lower estimation for the number of moves to reach the solved cube.
    log.debug('Preparing pruning table for the permutation of the edges in phase2.')
    Slice_URtoDF_Parity_Prun = load_cachetable('Slice_URtoDF_Parity_Prun')
    if not Slice_URtoDF_Parity_Prun:
        Slice_URtoDF_Parity_Prun = [-1] * (N_SLICE2 * N_URtoDF * N_PARITY // 2)  # new byte[N_SLICE2 * N_URtoDF * N_PARITY / 2]
        # Slice_URtoDF_Parity_Prun = [-1] * (N_SLICE2 * N_URtoDF * N_PARITY)  # new byte[N_SLICE2 * N_URtoDF * N_PARITY / 2]
        depth = 0
        setPruning(Slice_URtoDF_Parity_Prun, 0, 0)
        done = 1
        while (done != N_SLICE2 * N_URtoDF * N_PARITY):
            for i in range(N_SLICE2 * N_URtoDF * N_PARITY):
                parity = i % 2
                URtoDF = (i // 2) // N_SLICE2
                _slice = (i // 2) % N_SLICE2
                if (getPruning(Slice_URtoDF_Parity_Prun, i) == depth):
                    for j in range(18):
                        if j in (3, 5, 6, 8, 12, 14, 15, 17):
                            continue
                        else:
                            newSlice = FRtoBR_Move[_slice][j]
                            newURtoDF = URtoDF_Move[URtoDF][j]
                            newParity = parityMove[parity][j]
                            if (getPruning(Slice_URtoDF_Parity_Prun, (N_SLICE2 * newURtoDF + newSlice) * 2 + newParity) == 0x0f):
                                setPruning(
                                    Slice_URtoDF_Parity_Prun,
                                    (N_SLICE2 * newURtoDF + newSlice) * 2 + newParity,
                                    (depth + 1) & 0xff
                                )
                                done += 1
            depth += 1
        dump_cachetable(Slice_URtoDF_Parity_Prun, 'Slice_URtoDF_Parity_Prun')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Pruning table for the twist of the corners and the position (not permutation) of the UD-slice edges in phase1
    # The pruning table entries give a lower estimation for the number of moves to reach the H-subgroup.
    log.debug('Pruning table for the twist of the corners and the position (not permutation) of the UD-slice edges in phase1')
    Slice_Twist_Prun = load_cachetable('Slice_Twist_Prun')
    if not Slice_Twist_Prun:
        Slice_Twist_Prun = [-1] * (N_SLICE1 * N_TWIST // 2 + 1)  # new byte[N_SLICE1 * N_TWIST / 2 + 1]
        # Slice_Twist_Prun = [-1] * (N_SLICE1 * N_TWIST + 1)  # new byte[N_SLICE1 * N_TWIST / 2 + 1]
        depth = 0
        setPruning(Slice_Twist_Prun, 0, 0)
        done = 1
        while (done != N_SLICE1 * N_TWIST):
            for i in range(N_SLICE1 * N_TWIST):
                twist = i // N_SLICE1
                _slice = i % N_SLICE1
                if (getPruning(Slice_Twist_Prun, i) == depth):
                    for j in range(18):
                        newSlice = FRtoBR_Move[_slice * 24][j] // 24
                        newTwist = twistMove[twist][j]
                        if (getPruning(Slice_Twist_Prun, N_SLICE1 * newTwist + newSlice) == 0x0f):
                            setPruning(Slice_Twist_Prun, N_SLICE1 * newTwist + newSlice, (depth + 1) & 0xff)
                            done += 1

            depth += 1
        dump_cachetable(Slice_Twist_Prun, 'Slice_Twist_Prun')

    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # Pruning table for the flip of the edges and the position (not permutation) of the UD-slice edges in phase1
    # The pruning table entries give a lower estimation for the number of moves to reach the H-subgroup.
    log.debug('Pruning table for the flip of the edges and the position (not permutation) of the UD-slice edges in phase1')
    Slice_Flip_Prun = load_cachetable('Slice_Flip_Prun')
    if not Slice_Flip_Prun:
        Slice_Flip_Prun = [-1] * (N_SLICE1 * N_FLIP // 2)    # new byte[N_SLICE1 * N_FLIP / 2]
        # Slice_Flip_Prun = [-1] * (N_SLICE1 * N_FLIP)    # new byte[N_SLICE1 * N_FLIP / 2]
        depth = 0
        setPruning(Slice_Flip_Prun, 0, 0)
        done = 1
        while (done != N_SLICE1 * N_FLIP):
            for i in range(N_SLICE1 * N_FLIP):
                flip = i // N_SLICE1
                _slice = i % N_SLICE1
                if (getPruning(Slice_Flip_Prun, i) == depth):
                    for j in range(18):
                        newSlice = FRtoBR_Move[_slice * 24][j] // 24
                        newFlip = flipMove[flip][j]
                        if (getPruning(Slice_Flip_Prun, N_SLICE1 * newFlip + newSlice) == 0x0f):
                            setPruning(Slice_Flip_Prun, N_SLICE1 * newFlip + newSlice, (depth + 1) & 0xff)
                            done += 1
            depth += 1
        dump_cachetable(Slice_Flip_Prun, 'Slice_Flip_Prun')



from builtins import range

from .facelet import (
    U1, U2, U3, U4, U5, U6, U7, U8, U9, R1, R2, R3, R4, R5, R6, R7, R8, R9,
    F1, F2, F3, F4, F5, F6, F7, F8, F9, D1, D2, D3, D4, D5, D6, D7, D8, D9,
    L1, L2, L3, L4, L5, L6, L7, L8, L9, B1, B2, B3, B4, B5, B6, B7, B8, B9,
    facelet_values,
)
from .color import U, R, F, D, L, B, color_keys, colors
from .corner import URF, UFL, ULB, UBR, DFR, DLF, DBL, DRB, corner_values
from .edge import UR, UF, UL, UB, DR, DF, DL, DB, FR, FL, BL, BR, edge_values


class FaceCube(object):
    """Cube on the facelet level"""

    def __init__(self, cubeString="UUUUUUUUURRRRRRRRRFFFFFFFFFDDDDDDDDDLLLLLLLLLBBBBBBBBB"):
        self.f = []
        for c in cubeString:
            assert (c in colors)
            self.f.append(colors[c])

    # Map the corner positions to facelet positions. cornerFacelet[URF.ordinal()][0] e.g. gives the position of the
    # facelet in the URF corner position, which defines the orientation.<br>
    # cornerFacelet[URF.ordinal()][1] and cornerFacelet[URF.ordinal()][2] give the position of the other two facelets
    # of the URF corner (clockwise).
    cornerFacelet = [
        [ U9, R1, F3 ], [ U7, F1, L3 ], [ U1, L1, B3 ], [ U3, B1, R3 ],
        [ D3, F9, R7 ], [ D1, L9, F7 ], [ D7, B9, L7 ], [ D9, R9, B7 ],
    ]

    # Map the edge positions to facelet positions. edgeFacelet[UR.ordinal()][0] e.g. gives the position of the facelet in
    # the UR edge position, which defines the orientation.<br>
    # edgeFacelet[UR.ordinal()][1] gives the position of the other facelet
    edgeFacelet = [
        [ U6, R2 ], [ U8, F2 ], [ U4, L2 ], [ U2, B2 ], [ D6, R8 ], [ D2, F8 ],
        [ D4, L8 ], [ D8, B8 ], [ F6, R4 ], [ F4, L6 ], [ B6, L4 ], [ B4, R6 ],
    ]

    # Map the corner positions to facelet colors.
    cornerColor = [
        [ U, R, F ], [ U, F, L ], [ U, L, B ], [ U, B, R ],
        [ D, F, R ], [ D, L, F ], [ D, B, L ], [ D, R, B ],
    ]

    # Map the edge positions to facelet colors.
    edgeColor = [
        [ U, R ], [ U, F ], [ U, L ],
        [ U, B ], [ D, R ], [ D, F ],
        [ D, L ], [ D, B ], [ F, R ],
        [ F, L ], [ B, L ], [ B, R ],
    ]

    # Gives string representation of a facelet cube
    def to_String(self):
        return ''.join(color_keys[c] for c in self.f)

    # Gives CubieCube representation of a faceletcube
    def toCubieCube(self):
        from .cubiecube import CubieCube
        
        ccRet = CubieCube()
        for i in range(8):
            ccRet.cp[i] = URF   # invalidate corners
        for i in range(12):
            ccRet.ep[i] = UR    # and edges

        for i in corner_values:
            # get the colors of the cubie at corner i, starting with U/D
            for ori in range(3):
                if (self.f[self.cornerFacelet[i][ori]] == U
                        or self.f[self.cornerFacelet[i][ori]] == D):
                    break
            col1 = self.f[self.cornerFacelet[i][(ori + 1) % 3]]
            col2 = self.f[self.cornerFacelet[i][(ori + 2) % 3]]

            for j in corner_values:
                if (col1 == self.cornerColor[j][1]
                        and col2 == self.cornerColor[j][2]):
                    # in cornerposition i we have cornercubie j
                    ccRet.cp[i] = j
                    ccRet.co[i] = ori % 3
                    break

        for i in edge_values:
            for j in edge_values:
                if (self.f[self.edgeFacelet[i][0]] == self.edgeColor[j][0]
                        and self.f[self.edgeFacelet[i][1]] == self.edgeColor[j][1]):
                    ccRet.ep[i] = j
                    ccRet.eo[i] = 0
                    break

                if (self.f[self.edgeFacelet[i][0]] == self.edgeColor[j][1]
                        and self.f[self.edgeFacelet[i][1]] == self.edgeColor[j][0]):
                    ccRet.ep[i] = j
                    ccRet.eo[i] = 1
                    break

        return ccRet



import random
from builtins import range

from .facecube import FaceCube
from .cubiecube import CubieCube
from .coordcube import CoordCube
from .color import colors


def verify(s):
    """
    Check if the cube definition string s represents a solvable cube.

    @param s is the cube definition string , see {@link Facelet}
    @return 0: Cube is solvable<br>
            -1: There is not exactly one facelet of each colour<br>
            -2: Not all 12 edges exist exactly once<br>
            -3: Flip error: One edge has to be flipped<br>
            -4: Not all 8 corners exist exactly once<br>
            -5: Twist error: One corner has to be twisted<br>
            -6: Parity error: Two corners or two edges have to be exchanged
    """
    count = [0] * 6     # new int[6]
    try:
        for i in range(54):
            assert s[i] in colors
            count[colors[s[i]]] += 1
    except:
        return -1

    for i in range(6):
        if count[i] != 9:
            return -1

    fc = FaceCube(s)
    cc = fc.toCubieCube()

    return cc.verify()


def randomCube():
    """
    Generates a random cube.
    @return A random cube in the string representation. Each cube of the cube space has the same probability.
    """
    cc = CubieCube()
    cc.setFlip(random.randint(0, CoordCube.N_FLIP - 1))
    cc.setTwist(random.randint(0, CoordCube.N_TWIST - 1))
    while True:
        cc.setURFtoDLB(random.randint(0, CoordCube.N_URFtoDLB - 1))
        cc.setURtoBR(random.randint(0, CoordCube.N_URtoBR - 1))

        if (cc.edgeParity() ^ cc.cornerParity()) == 0:
            break
    fc = cc.toFaceCube()
    return fc.to_String()


def randomLastLayerCube():
    """
    Generates a cube with a random last layer.
    @return A cube with a random last layer and otherwise solved facelets in the string representation.
    """
    cc = CubieCube()
    cc.setFlip(random.choice([0, 24, 40, 48, 72, 80, 96, 120]))
    cc.setTwist(random.randint(0, 26))
    while True:
        perms = [0, 624, 3744, 3840, 4344, 4440, 26064, 26160, 26664, 26760,
                 27360, 27984, 30384, 30480, 30984, 31080, 31680, 32304, 35304,
                 35400, 36000, 36624, 39744, 39840]
        cc.setURFtoDLB(random.choice(perms))
        cc.setURtoBR(random.choice(perms))

        if (cc.edgeParity() ^ cc.cornerParity()) == 0:
            break
    fc = cc.toFaceCube()
    return fc.to_String()



import copy
from builtins import range

from .corner import URF, UFL, ULB, UBR, DFR, DLF, DBL, DRB, corner_values
from .edge import UR, UF, UL, UB, DR, DF, DL, DB, FR, FL, BL, BR, edge_values
from .facecube import FaceCube


# n choose k
def Cnk(n, k):
    if n < k:
        return 0
    if k > n // 2:
        k = n - k
    s = 1
    i = n
    j = 1
    while i != n - k:
        s *= i
        s //= j
        i -= 1
        j += 1
    return s


def rotateLeft(arr, l, r):
    """Left rotation of all array elements between l and r"""
    temp = arr[l]
    for i in range(l, r):
        arr[i] = arr[i + 1]
    arr[r] = temp


def rotateRight(arr, l, r):
    """Right rotation of all array elements between l and r"""
    temp = arr[r]
    for i in range(r, l, -1):
        arr[i] = arr[i - 1]
    arr[l] = temp


def getURtoDF(idx1, idx2):
    """Permutation of the six edges UR,UF,UL,UB,DR,DF"""
    a = CubieCube()
    b = CubieCube()
    a.setURtoUL(idx1)
    b.setUBtoDF(idx2)
    for i in range(8):
        if a.ep[i] != BR:
            if b.ep[i] != BR:   # collision
                return -1
            else:
                b.ep[i] = a.ep[i]
    return b.getURtoDF()


class CubieCube(object):
    """Cube on the cubie level"""

    # initialize to Id-Cube

    def __init__(self, cp=None, co=None, ep=None, eo=None):
        # corner permutation
        self.cp = copy.copy(cp) if cp else [URF, UFL, ULB, UBR, DFR, DLF, DBL, DRB]

        # corner orientation
        self.co = copy.copy(co) if co else [0, 0, 0, 0, 0, 0, 0, 0]

        # edge permutation
        self.ep = copy.copy(ep) if ep else [UR, UF, UL, UB, DR, DF, DL, DB, FR, FL, BL, BR]

        # edge orientation
        self.eo = copy.copy(eo) if eo else [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    def toFaceCube(self):
        """return cube in facelet representation"""

        fcRet = FaceCube()
        for i in corner_values:
            j = self.cp[i]  # cornercubie with index j is at cornerposition with index i
            ori = self.co[i]   # Orientation of this cubie
            for n in range(3):
                _butya = FaceCube.cornerFacelet[i][(n + ori) % 3]
                fcRet.f[_butya] = FaceCube.cornerColor[j][n]
        for i in edge_values:
            ori = self.eo[i]   # Orientation of this cubie
            for n in range(2):
                _butya = FaceCube.edgeFacelet[i][(n + ori) % 2]
                fcRet.f[_butya] = FaceCube.edgeColor[self.ep[i]][n]
        return fcRet

    def cornerMultiply(self, b):
        """
        Multiply this CubieCube with another cubiecube b, restricted to the corners.<br>
        Because we also describe reflections of the whole cube by permutations, we get a complication with the corners. The
        orientations of mirrored corners are described by the numbers 3, 4 and 5. The composition of the orientations
        cannot
        be computed by addition modulo three in the cyclic group C3 any more. Instead the rules below give an addition in
        the dihedral group D3 with 6 elements.<br>

        NOTE: Because we do not use symmetry reductions and hence no mirrored cubes in this simple implementation of the
        Two-Phase-Algorithm, some code is not necessary here.

        b - CubieCube instance
        """

        cPerm = []  # new Corner[8]
        cOri = []   # new byte[8]
        for i in corner_values:
            cPerm.append(self.cp[b.cp[i]])

            oriA = self.co[b.cp[i]]
            oriB = b.co[i]
            ori = 0

            if oriA < 3 and oriB < 3:  # if both cubes are regular cubes...
                ori = (oriA + oriB) & 0xff   # just do an addition modulo 3 here
                if ori >= 3:
                    ori -= 3    # the composition is a regular cube

            # +++++++++++++++++++++not used in this implementation +++++++++++++
            elif oriA < 3 and oriB >= 3:    # if cube b is in a mirrored
                # state...
                ori = (oriA + oriB) & 0xff
                if ori >= 6:
                    ori -= 3    # the composition is a mirrored cube
            elif oriA >= 3 and oriB < 3:    # if cube a is an a mirrored
                # state...
                ori = (oriA - oriB) & 0xff
                if ori < 3:
                    ori += 3    # the composition is a mirrored cube
            elif oriA >= 3 and oriB >= 3:   # if both cubes are in mirrored
                # states...
                ori = (oriA - oriB) & 0xff
                if ori < 0:
                    ori += 3    # the composition is a regular cube
            # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            cOri.append(ori)

        for i in corner_values:
            self.cp[i] = cPerm[i]
            self.co[i] = cOri[i]

    def edgeMultiply(self, b):
        """
        Multiply this CubieCube with another cubiecube b, restricted to the edges.

        b - CubieCube instance
        """

        ePerm = []  # new Edge[12]
        eOri = []   # new byte[12]
        for i in edge_values:
            _ = b.ep[i]
            ePerm.append(self.ep[_])
            eOri.append(((b.eo[i] + self.eo[_]) % 2) & 0xff)

        for i in edge_values:
            self.ep[i] = ePerm[i]
            self.eo[i] = eOri[i]

    def multiply(self, b):
        """
        Multiply this CubieCube with another CubieCube b.

        b - CubieCube instance
        """

        self.cornerMultiply(b)
        self.edgeMultiply(b)

    def invCubieCube(self, c):
        """
        Compute the inverse CubieCube

        c - CubieCube instance
        """

        for i in edge_values:
            c.ep[self.ep[i]] = i
        for i in edge_values:
            c.eo[i] = self.eo[c.ep[i]]
        for i in corner_values:
            c.cp[self.cp[i]] = i
        for i in corner_values:
            ori = self.co[c.cp[i]]
            if ori >= 3:
                # Just for completeness. We do not invert mirrored cubes in the program.
                c.co[i] = ori
            else:
                # the standard case
                c.co[i] = -ori
                if c.co[i] < 0:
                    c.co[i] += 3

    # ********************* Get and set coordinates ***************************

    def getTwist(self):
        """return the twist of the 8 corners. 0 <= twist < 3^7"""

        ret = 0
        for i in range(URF, DRB):
            ret = (3 * ret + self.co[i]) & 0xffff
        return ret

    def setTwist(self, twist):
        twistParity = 0
        for i in range(DRB - 1, URF - 1, -1):
            self.co[i] = (twist % 3) & 0xff
            twistParity += self.co[i]
            twist //= 3
        self.co[DRB] = ((3 - twistParity % 3) % 3) & 0xff

    def getFlip(self):
        """return the flip of the 12 edges. 0<= flip < 2^11"""

        ret = 0
        for i in range(UR, BR):
            ret = (2 * ret + self.eo[i]) & 0xffff
        return ret

    def setFlip(self, flip):
        flipParity = 0
        for i in range(BR - 1, UR - 1, -1):
            self.eo[i] = (flip % 2) & 0xff
            flipParity += self.eo[i]
            flip //= 2
        self.eo[BR] = ((2 - flipParity % 2) % 2) & 0xff

    def cornerParity(self):
        """Parity of the corner permutation"""
        s = 0
        for i in range(DRB, URF, -1):
            for j in range(i - 1, URF - 1, -1):
                if self.cp[j] > self.cp[i]:
                    s += 1
        return (s % 2) & 0xffff

    def edgeParity(self):
        """Parity of the edges permutation. Parity of corners and edges are the same if the cube is solvable."""
        s = 0
        for i in range(BR, UR, -1):
            for j in range(i - 1, UR - 1, -1):
                if self.ep[j] > self.ep[i]:
                    s += 1
        return (s % 2) & 0xffff

    def getFRtoBR(self):
        """permutation of the UD-slice edges FR,FL,BL and BR"""
        a = 0
        x = 0
        edge4 = [None] * 4  # new Edge[4]
        # compute the index a < (12 choose 4) and the permutation array perm.
        for j in range(BR, UR - 1, -1):
            if (FR <= self.ep[j] and self.ep[j] <= BR):
                a += Cnk(11 - j, x + 1)
                edge4[3 - x] = self.ep[j]
                x += 1

        b = 0
        for j in range(3, 0, -1):  # compute the index b < 4! for the permutation in perm
            k = 0
            while (edge4[j] != j + 8):
                rotateLeft(edge4, 0, j)
                k += 1
            b = (j + 1) * b + k
        return (24 * a + b) & 0xffff

    def setFRtoBR(self, idx):
        sliceEdge = [FR, FL, BL, BR]
        otherEdge = [UR, UF, UL, UB, DR, DF, DL, DB]
        b = idx % 24   # Permutation
        a = idx // 24   # Combination
        for i in edge_values:
            self.ep[i] = DB     # Use UR to invalidate all edges

        for j in range(1, 4):  # generate permutation from index b
            k = b % (j + 1)
            b //= j + 1

            while k > 0:    # while (k-- > 0) #????????????????
                k -= 1
                rotateRight(sliceEdge, 0, j)

        x = 3   # generate combination and set slice edges
        for j in range(UR, BR + 1):
            if a - Cnk(11 - j, x + 1) >= 0:
                self.ep[j] = sliceEdge[3 - x]
                a -= Cnk(11 - j, x + 1)
                x -= 1
        x = 0   # set the remaining edges UR..DB
        for j in range(UR, BR + 1):
            if self.ep[j] == DB:
                self.ep[j] = otherEdge[x]
                x += 1

    def getURFtoDLF(self):
        """Permutation of all corners except DBL and DRB"""
        a = 0
        x = 0
        corner6 = []    # new Corner[6]
        # compute the index a < (8 choose 6) and the corner permutation.
        for j in range(URF, DRB + 1):
            if self.cp[j] <= DLF:
                a += Cnk(j, x + 1)
                corner6.append(self.cp[j])
                x += 1

        b = 0
        for j in range(5, 0, -1):   # compute the index b < 6! for the
            # permutation in corner6
            k = 0
            while corner6[j] != j:
                rotateLeft(corner6, 0, j)
                k += 1
            b = (j + 1) * b + k
        return (720 * a + b) & 0xffff

    def setURFtoDLF(self, idx):
        corner6 = [URF, UFL, ULB, UBR, DFR, DLF]
        otherCorner = [DBL, DRB]
        b = idx % 720  # Permutation
        a = idx // 720  # Combination
        for i in corner_values:
            self.cp[i] = DRB    # Use DRB to invalidate all corners

        for j in range(1, 6):   # generate permutation from index b
            k = b % (j + 1)
            b //= j + 1
            while k > 0:
                k -= 1
                rotateRight(corner6, 0, j)
        x = 5
        # generate combination and set corners
        for j in range(DRB, -1, -1):
            if a - Cnk(j, x + 1) >= 0:
                self.cp[j] = corner6[x]
                a -= Cnk(j, x + 1)
                x -= 1
        x = 0
        for j in range(URF, DRB + 1):
            if self.cp[j] == DRB:
                self.cp[j] = otherCorner[x]
                x += 1

    def getURtoDF(self):
        """Permutation of the six edges UR,UF,UL,UB,DR,DF."""
        a = 0
        x = 0
        edge6 = []  # new Edge[6]
        # compute the index a < (12 choose 6) and the edge permutation.
        for j in range(UR, BR + 1):
            if self.ep[j] <= DF:
                a += Cnk(j, x + 1)
                edge6.append(self.ep[j])
                x += 1

        b = 0
        for j in range(5, 0, -1):  # compute the index b < 6! for the permutation in edge6
            k = 0
            while edge6[j] != j:
                rotateLeft(edge6, 0, j)
                k += 1
            b = (j + 1) * b + k
        return 720 * a + b

    def setURtoDF(self, idx):
        edge6 = [UR, UF, UL, UB, DR, DF]
        otherEdge = [DL, DB, FR, FL, BL, BR]
        b = idx % 720  # Permutation
        a = idx // 720  # Combination
        for i in edge_values:
            self.ep[i] = BR     # Use BR to invalidate all edges

        for j in range(1, 6):   # generate permutation from index b
            k = b % (j + 1)
            b //= j + 1
            while k > 0:
                k -= 1
                rotateRight(edge6, 0, j)
        x = 5
        # generate combination and set edges
        for j in range(BR, -1, -1):
            if a - Cnk(j, x + 1) >= 0:
                self.ep[j] = edge6[x]
                a -= Cnk(j, x + 1)
                x -= 1
        x = 0
        # set the remaining edges DL..BR
        for j in range(UR, BR + 1):
            if self.ep[j] == BR:
                self.ep[j] = otherEdge[x]
                x += 1

    def getURtoUL(self):
        """Permutation of the three edges UR,UF,UL"""
        a = 0
        x = 0
        edge3 = []  # new Edge[3]
        # compute the index a < (12 choose 3) and the edge permutation.
        for j in range(UR, BR + 1):
            if self.ep[j] <= UL:
                a += Cnk(j, x + 1)
                edge3.append(self.ep[j])
                x += 1

        b = 0
        for j in range(2, 0, -1):  # compute the index b < 3! for the permutation in edge3
            k = 0
            while edge3[j] != j:
                rotateLeft(edge3, 0, j)
                k += 1
            b = (j + 1) * b + k
        return (6 * a + b) & 0xffff

    def setURtoUL(self, idx):
        edge3 = [UR, UF, UL]
        b = idx % 6    # Permutation
        a = idx // 6    # Combination
        for i in edge_values:
            self.ep[i] = BR    # Use BR to invalidate all edges

        for j in range(1, 3):   # generate permutation from index b
            k = b % (j + 1)
            b //= j + 1
            while k > 0:
                k -= 1
                rotateRight(edge3, 0, j)
        x = 2  # generate combination and set edges
        for j in range(BR, -1, -1):
            if a - Cnk(j, x + 1) >= 0:
                self.ep[j] = edge3[x]
                a -= Cnk(j, x + 1)
                x -= 1

    def getUBtoDF(self):
        """Permutation of the three edges UB,DR,DF"""
        a = 0
        x = 0
        edge3 = []  # new Edge[3]
        # compute the index a < (12 choose 3) and the edge permutation.
        for j in range(UR, BR + 1):
            if UB <= self.ep[j] and self.ep[j] <= DF:
                a += Cnk(j, x + 1)
                edge3.append(self.ep[j])
                x += 1

        b = 0
        for j in range(2, 0, -1):  # compute the index b < 3! for the permutation in edge3
            k = 0
            while edge3[j] != UB + j:
                rotateLeft(edge3, 0, j)
                k += 1
            b = (j + 1) * b + k
        return (6 * a + b) & 0xffff

    def setUBtoDF(self, idx):
        edge3 = [UB, DR, DF]
        b = idx % 6    # Permutation
        a = idx // 6    # Combination
        for i in edge_values:
            self.ep[i] = BR     # Use BR to invalidate all edges

        for j in range(1, 3):
            # generate permutation from index b
            k = b % (j + 1)
            b //= j + 1
            while k > 0:
                k -= 1
                rotateRight(edge3, 0, j)
        x = 2
        # generate combination and set edges
        for j in range(BR, -1, -1):
            if a - Cnk(j, x + 1) >= 0:
                self.ep[j] = edge3[x]
                a -= Cnk(j, x + 1)
                x -= 1

    def getURFtoDLB(self):
        perm = copy.copy(self.cp)
        b = 0
        for j in range(7, 0, -1):  # compute the index b < 8! for the permutation in perm
            k = 0
            while perm[j] != j:
                rotateLeft(perm, 0, j)
                k += 1
            b = (j + 1) * b + k
        return b

    def setURFtoDLB(self, idx):
        perm = [URF, UFL, ULB, UBR, DFR, DLF, DBL, DRB]
        for j in range(1, 8):
            k = idx % (j + 1)
            idx //= j + 1
            while k > 0:
                k -= 1
                rotateRight(perm, 0, j)
        x = 7
        # set corners
        for j in range(7, -1, -1):
            self.cp[j] = perm[x]
            x -= 1

    def getURtoBR(self):
        perm = copy.copy(self.ep)
        b = 0
        for j in range(11, 0, -1):     # compute the index b < 12! for the permutation in perm
            k = 0
            while perm[j] != j:
                rotateLeft(perm, 0, j)
                k += 1
            b = (j + 1) * b + k
        return b

    def setURtoBR(self, idx):
        perm = [UR, UF, UL, UB, DR, DF, DL, DB, FR, FL, BL, BR]
        for j in range(1, 12):
            k = idx % (j + 1)
            idx //= j + 1
            while k > 0:
                k -= 1
                rotateRight(perm, 0, j)
        x = 11  # set edges
        for j in range(11, -1, -1):
            self.ep[j] = perm[x]
            x -= 1

    def verify(self):
        """
        Check a cubiecube for solvability. Return the error code.
        0: Cube is solvable
        -2: Not all 12 edges exist exactly once
        -3: Flip error: One edge has to be flipped
        -4: Not all corners exist exactly once
        -5: Twist error: One corner has to be twisted
        -6: Parity error: Two corners ore two edges have to be exchanged
        """

        sum = 0
        edgeCount = [0] * 12  # new int[12]
        for i in edge_values:
            edgeCount[self.ep[i]] += 1
        for i in range(12):
            if edgeCount[i] != 1:
                return -2

        for i in range(12):
            sum += self.eo[i]
        if sum % 2 != 0:
            return -3

        cornerCount = [0] * 8   # new int[8]
        for i in corner_values:
            cornerCount[self.cp[i]] += 1
        for i in range(8):
            if cornerCount[i] != 1:
                return -4   # missing corners

        sum = 0
        for i in range(8):
            sum += self.co[i]
        if sum % 3 != 0:
            return -5   # twisted corner

        if (self.edgeParity() ^ self.cornerParity()) != 0:
            return -6   # parity error

        return 0    # cube ok


# ************************ Moves on the cubie level ****************************

cpU = [UBR, URF, UFL, ULB, DFR, DLF, DBL, DRB]
coU = [0, 0, 0, 0, 0, 0, 0, 0]
epU = [UB, UR, UF, UL, DR, DF, DL, DB, FR, FL, BL, BR]
eoU = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

cpR = [DFR, UFL, ULB, URF, DRB, DLF, DBL, UBR]
coR = [2, 0, 0, 1, 1, 0, 0, 2]
epR = [FR, UF, UL, UB, BR, DF, DL, DB, DR, FL, BL, UR]
eoR = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
cpF = [UFL, DLF, ULB, UBR, URF, DFR, DBL, DRB]
coF = [1, 2, 0, 0, 2, 1, 0, 0]
epF = [UR, FL, UL, UB, DR, FR, DL, DB, UF, DF, BL, BR]
eoF = [0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0]
cpD = [URF, UFL, ULB, UBR, DLF, DBL, DRB, DFR]
coD = [0, 0, 0, 0, 0, 0, 0, 0]
epD = [UR, UF, UL, UB, DF, DL, DB, DR, FR, FL, BL, BR]
eoD = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
cpL = [URF, ULB, DBL, UBR, DFR, UFL, DLF, DRB]
coL = [0, 1, 2, 0, 0, 2, 1, 0]
epL = [UR, UF, BL, UB, DR, DF, FL, DB, FR, UL, DL, BR]
eoL = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
cpB = [URF, UFL, UBR, DRB, DFR, DLF, ULB, DBL]
coB = [0, 0, 1, 2, 0, 0, 2, 1]
epB = [UR, UF, UL, BR, DR, DF, DL, BL, FR, FL, UB, DB]
eoB = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1]

# this CubieCube array represents the 6 basic cube moves
moveCube = [
    CubieCube(cp=cpU, co=coU, ep=epU, eo=eoU),
    CubieCube(cp=cpR, co=coR, ep=epR, eo=eoR),
    CubieCube(cp=cpF, co=coF, ep=epF, eo=eoF),
    CubieCube(cp=cpD, co=coD, ep=epD, eo=eoD),
    CubieCube(cp=cpL, co=coL, ep=epL, eo=eoL),
    CubieCube(cp=cpB, co=coB, ep=epB, eo=eoB),
]



import os
import io

from setuptools import setup

long_description = io.open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8').read()

setup(
    name='kociemba',
    version='1.2.1',
    description='Python/C implementation of Herbert Kociemba\'s Two-Phase algorithm for solving Rubik\'s Cube',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='kociemba rubik cube solver twophase',
    url='https://github.com/muodov/kociemba',
    author='muodov',
    author_email='muodov@gmail.com',
    license='GPLv2',
    packages=['kociemba'],
    package_data={
        '': [
            'cprunetables/*',
            'pykociemba/*.py',
            'pykociemba/prunetables/*',
            'ckociemba/include/*.h',
            'ckociemba/*.c'],
    },
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
    ),
    entry_points = {
        'console_scripts': ['kociemba=kociemba.command_line:main'],
    },
    setup_requires=['pytest-runner', "cffi>=1.0.0"],
    tests_require=['pytest', ],
    zip_safe=False,
    cffi_modules=["kociemba/build_ckociemba.py:ffi"],
    install_requires=["cffi>=1.0.0", 'future'],
)
